/*****************************************************************************/
/* Kernel SOCKS Bouncer (Loadable Kernel Module) for 2.6.x kernels [KSB26]   */
/* (c) 2004 Paolo Ardoino <ardoino.gnu@disi.unige.it>                        */
/*****************************************************************************/
/*									     */
/* This program is free software; you can redistribute it and/or modify	     */
/* it under the terms of the GNU General Public License as published by	     */
/* the Free Software Foundation; either version 2 of the License, or	     */
/* (at your option) any later version.					     */
/* This program is distributed in the hope that it will be useful,	     */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	     */
/* GNU General Public License for more details.				     */
/*									     */
/* You should have received a copy of the GNU General Public License	     */
/* along with this program; if not, write to the Free Software		     */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*****************************************************************************/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/moduleparam.h>
#include <asm/unistd.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/preempt.h>
#include "kpatch/ksb26.h"

#define KSB26_DEV_NAME "/dev/ksb26"
#define KSB26_IOCTL_BUFLEN 4096

#define DEBUG 1

extern char ksb26;
extern unsigned int ksb26_nsocks;
extern unsigned int ksb26_naddrs;
extern struct ksb26_sockscell *ksb26_sl;
extern struct ksb26_addrs *ksb26_al;

static int ksb26_major;
static char ksb26_ioctl_msg[KSB26_IOCTL_BUFLEN];
static int ksb26_dev_bool = 0;
static struct ksb26_sockscell *ksb26_slaux;
static struct ksb26_addrs *ksb26_alaux;

static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static ssize_t device_write(struct file *, const char *, size_t, loff_t *);

static struct file_operations fops = {
	.write = device_write,
	.open = device_open,
	.release = device_release
};

static int device_open(struct inode *inode, struct file *file)
{
	if(ksb26_dev_bool == 1) 
		return -EBUSY;	
	ksb26_dev_bool++;
	try_module_get(THIS_MODULE);
	return 0;
}

static int device_release(struct inode *inode, struct file *file)
{
	ksb26_dev_bool--;
	module_put(THIS_MODULE);
	return 0;
}

static ssize_t device_write(struct file *filp, const char *buffer, size_t lenght, loff_t *offset)
{
	int bytes_write = 0;
	char *aux;
	
	for(bytes_write = 0; bytes_write < lenght && bytes_write < KSB26_IOCTL_BUFLEN; bytes_write++)
		get_user(ksb26_ioctl_msg[bytes_write], buffer + bytes_write);
	ksb26_ioctl_msg[lenght - 1] = '\0';
	aux = ksb26_ioctl_msg;
	while(aux && *(aux+1)) {
		if(aux[0] == 's') {
			aux++;
			ksb26_slaux->host = kmalloc(18, GFP_KERNEL);
			if(sscanf(aux, "%s %d", ksb26_slaux->host, &ksb26_slaux->port) == 2) {
#ifdef DEBUG
				printk("[%s] %s:%d socks5 added.\n", MODNAME, ksb26_slaux->host, ksb26_slaux->port);
#endif
				ksb26_slaux->next = (struct ksb26_sockscell *)kmalloc(sizeof(struct ksb26_sockscell), GFP_KERNEL);
				ksb26_slaux = ksb26_slaux->next;
				ksb26_nsocks++;
#ifdef DEBUG
				printk("[%s] %d socks in list.\n", MODNAME, ksb26_nsocks);
#endif	
			}
		} else if(aux[0] == 'a') {
			aux++;
			ksb26_alaux->addr = in_aton(aux);
#ifdef DEBUG
			printk("[%s] Addr = %s [%lu] added.\n", MODNAME, aux, ksb26_alaux->addr);
#endif
			ksb26_alaux->next = (struct ksb26_addrs *)kmalloc(sizeof(struct ksb26_addrs), GFP_KERNEL);
			ksb26_alaux = ksb26_alaux->next;
			ksb26_naddrs++;
#ifdef DEBUG
			printk("[%s] %d addrs in list.\n", MODNAME, ksb26_naddrs);
#endif
		
		} else
			aux++;
	}
	return bytes_write;
}

static int __init modinit(void)
{
	printk("[%s v%s] MODULE LOADED\n", MODNAME, KSB26_VERSION);
	if((ksb26_major = register_chrdev(0, KSB26_DEV_NAME, &fops)) < 0) {
		printk("[%s] Cannot open device %s with %d major.\n", MODNAME, KSB26_DEV_NAME, ksb26_major);
		return -ENODEV;
	}
	printk("[%s] Device %s opened.\n", MODNAME, KSB26_DEV_NAME);
	printk("[%s] Major device number = %d.\n", MODNAME, ksb26_major);
	printk("[%s] mknod %s c %d 0\n", MODNAME, KSB26_DEV_NAME, ksb26_major);
	ksb26_sl = (struct ksb26_sockscell *)kmalloc(sizeof(struct ksb26_sockscell), GFP_KERNEL);
	ksb26_slaux = ksb26_sl;
	ksb26_al = (struct ksb26_addrs *)kmalloc(sizeof(struct ksb26_addrs), GFP_KERNEL);
	ksb26_alaux = ksb26_al;
	ksb26_nsocks = 0;
	ksb26_naddrs = 0;
	ksb26 = 1;
	printk("[%s] Hijacking SYS_CONNECT syscall.\n", MODNAME);
	return 0;
}

static void __exit modcleanup(void)
{
	ksb26 = 0;
	ksb26_nsocks = 0;
	if(unregister_chrdev(ksb26_major, KSB26_DEV_NAME) < 0)
		printk("[%s] Cannot unregister device %s.\n", MODNAME, KSB26_DEV_NAME);
	printk("[%s] MODULE CLEANED\n", MODNAME);
	return;
}

module_init(modinit);
module_exit(modcleanup);

MODULE_AUTHOR("Paolo Ardoino");
MODULE_DESCRIPTION("Kernel SOCKS Bouncer for 2.6.x kernels\n");
MODULE_LICENSE("GPL");
